@@ -1,4 +1,4 @@
-from importlib import find_loader as module_exists
+from importlib.util import find_spec as module_exists
 
 avail_solvers = []
 if module_exists('swiglpk'):
@@ -11,16 +11,15 @@ if module_exists('pyscipopt'):
     avail_solvers += ['scip']
     
 from .names import *
-from .strainDesignModule import *
-from .strainDesignSolution import *
-from .indicatorConstraints import *
 from .solver_interface import *
+from .indicatorConstraints import *
 from .pool import *
 from .efmtool import *
 from .parse_constr import *
+from .lptools import *
+from .strainDesignModule import *
+from .strainDesignSolution import *
 from .strainDesignMILPBuilder import *
 from .strainDesignMILP import *
 from .strainDesigner import *
-from .fba import *
-from .fva import *
-from .compute_strain_designs import *
\ No newline at end of file
+from .compute_strain_designs import *
